#ifndef _HERMIT_TARGET_TAGS_H
#define _HERMIT_TARGET_TAGS_H

void hermit_entry();
int hermit_linux_cmdfunc(int argc, char *argv[]);

#define SRAM_START              0x60000000
// #define SRAM_SIZE               0x0000c000
#define DRAM1_START             0xC0000000
#define DRAM1_SIZE              0x01000000



#define IO_START        0x80000000
#define IO(offset)      (IO_START + (offset))
#define IO_BYTE(offset) (*(volatile unsigned char *)(IO_START + (offset)))
#define IO_WORD(offset) (*(volatile unsigned long *)(IO_START + (offset)))
// #define LEDFLSH 0x22C0  /* LED Flash control register ---------------------- */
#define IO_LEDFLSH      IO_WORD(LEDFLSH)

#define LINUX_PARAM_SIZE        (0x00000f00)
#define LINUX_PARAM_ADDRESS     (DRAM1_START + 0x00000100)
#define LINUX_LOAD_ADDRESS      (DRAM1_START + 0x00028000)
#define INITRD_SIZE             (0x00300000)
#define INITRD_LOAD_ADDRESS     (DRAM1_START + 0x00300000)


/* for setting the root device */
#define MINORBITS       8
#define MKDEV(ma,mi)    (((ma) << MINORBITS) | (mi))
/* #define RAMDISK_MAJOR   1     major for "RAM disk" */
/* #define RAMDISK0_MINOR  0     b 1 0 == /dev/ram0 */
/*#define ROM_FLASH_MAJOR       31       major for "ROM/flash memory card" */
/*#define FLASH0_MINOR  16              b 31 16 == /dev/flash0 */
/*#define FLASH1_MINOR  17               b 31 17 == /dev/flash1 */
#define MTD_MAJOR       31
#define MTD_MINOR       3               /* i boot on /dev/mtd3 */
#define COMMAND_LINE_SIZE 1024

 
/* lifted from linux/include/asm-arm/setup.h */

// #include <target/htypes.h>

//typedef uint8_t  u8;
//typedef uint16_t u16;
//typedef uint32_t u32;


#define ATAG_NONE	0x00000000

struct tag_header {
	u32 size;
	u32 tag;
};

#define ATAG_CORE	0x54410001

struct tag_core {
	u32 flags;		/* bit 0 = read-only */
	u32 pagesize;
	u32 rootdev;
};

#define ATAG_MEM		0x54410002

struct tag_mem32 {
	u32	size;
	u32	start;
};

#define ATAG_VIDEOTEXT	0x54410003

struct tag_videotext {
	u8		x;
	u8		y;
	u16		video_page;
	u8		video_mode;
	u8		video_cols;
	u16		video_ega_bx;
	u8		video_lines;
	u8		video_isvga;
	u16		video_points;
};

#define ATAG_RAMDISK	0x54410004

struct tag_ramdisk {
	u32 flags;		/* b0 = load, b1 = prompt */
	u32 size;
	u32 start;
};

#define ATAG_INITRD	0x54410005

struct tag_initrd {
	u32 start;
	u32 size;
};

#define ATAG_SERIAL	0x54410006

struct tag_serialnr {
	u32 low;
	u32 high;
};

#define ATAG_REVISION	0x54410007

struct tag_revision {
	u32 rev;
};

#define ATAG_VIDEOLFB	0x54410008

struct tag_videolfb {
	u16		lfb_width;
	u16		lfb_height;
	u16		lfb_depth;
	u16		lfb_linelength;
	u32		lfb_base;
	u32		lfb_size;
	u8		red_size;
	u8		red_pos;
	u8		green_size;
	u8		green_pos;
	u8		blue_size;
	u8		blue_pos;
	u8		rsvd_size;
	u8		rsvd_pos;
};

#define ATAG_CMDLINE	0x54410009

struct tag_cmdline {
	char	cmdline[1];
};

#define ATAG_ACORN	0x41000101

struct tag_acorn {
	u32 memc_control_reg;
	u32 vram_pages;
	u8 sounddefault;
	u8 adfsdrives;
};

#define ATAG_MEMCLK	0x41000402

struct tag_memclk {
	u32 fmemclk;
};

struct tag {
	struct tag_header hdr;
	union {
		struct tag_core		core;
		struct tag_mem32	mem;
		struct tag_videotext	videotext;
		struct tag_ramdisk	ramdisk;
		struct tag_initrd	initrd;
		struct tag_serialnr	serialnr;
		struct tag_revision	revision;
		struct tag_videolfb	videolfb;
		struct tag_cmdline	cmdline;

		/*
		 * Acorn specific
		 */
		struct tag_acorn	acorn;

		/*
		 * DC21285 specific
		 */
		struct tag_memclk	memclk;
	} u;
};

struct tagtable {
	u32 tag;
	int (*parse)(const struct tag *);
};

#define tag_member_present(tag,member)		 ((unsigned long)(&((struct tag *)0L)->member + 1) <= (tag)->hdr.size * 4)

#define tag_next(t)	((struct tag *)((u32 *)(t) + (t)->hdr.size))
#define tag_size(type)	((sizeof(struct tag_header) + sizeof(struct type)) >> 2)

#endif /* _HERMIT_TARGET_TAGS_H */
